/*

Files     	: _4RGB_LED_Matrix_TheSnake.ino
Version 	: 0.1
Date     	: 2012/12/06
Description	: This code is not optimized.
History		: 

    1. Date		: 2012/11/12
       Author		: John Zhong@SimplyTronics
       Modification	: Create   

    1. Date		: 2012/12/06
       Author		: John Zhong@SimplyTronics
       Modification	: Only a few changes to the marco
                          to make it fix into the 4 units of RGB LED Matrix Module.
                          Change the font size to 0508.
   	   

This source code is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <RGB_LED_Matrix.h>

/* connect the module to pin 1 Serial Port 'Tx' */
/*#define  RGB_LED_MATRIX_SIO (1) */


#define SIGLE_DIRECTION_LIST_LENGTH (32)  // The length of the single direction list
#define SNAKE_LENGTH_MIN (5)              // the original length of the snake
#define SNAKE_LENGTH_MAX (SIGLE_DIRECTION_LIST_LENGTH-1) // the Max length of the snake

#define BUTTON_MIDDLE (2)                              // 5-position switch
#define BUTTON_UP   (3)
#define BUTTON_LEFT (4)
#define BUTTON_DOWN (5)
#define BUTTON_RIGHT (6)

#define AREA_BOUND_START_X (0)                        // The snake moving area 
#define AREA_BOUND_START_Y (0)
#define AREA_BOUND_END_X (15)
#define AREA_BOUND_END_Y (15)

#define  RGB_LED_MATRIX_COLOR_BLACK (0x00000000)
#define  RGB_LED_MATRIX_COLOR_WHITE (0x00FFFFFF)
#define  RGB_LED_MATRIX_COLOR_RED (0x000000FF)
#define  RGB_LED_MATRIX_COLOR_GREEN (0x0000FF00)
#define  RGB_LED_MATRIX_COLOR_BLUE (0x00FF0000)

/* C++ object */
RGB_LED_Matrix matrix;


typedef struct snake_status_st                       // The snake status
{   
    unsigned char x_or_y;
    signed char move_step;
    unsigned char life;
} snake_status_struct;


typedef struct snake_st                            // the body of the snake, for each part
{
    unsigned char x;
    unsigned char y;
    struct snake_st *next;
} snake_struct;

/*  Global variables */
snake_status_struct snake_status;

snake_struct snake[SIGLE_DIRECTION_LIST_LENGTH];
snake_struct *p_empty_snake;                  // Point to empty list 
snake_struct *p_food;                         // Point to the food 
snake_struct *p_head_snake;                   // Point to the head of the snake 
snake_struct *p_current_snake;                // Point to current part(the body)      
snake_struct *p_tail_snake;                   // Point to the tail of the snake 

snake_struct temp0_snake;                      // two run-time variables
snake_struct temp1_snake;

unsigned char flag_got_food;                    // The flag is set when the snake got the food
unsigned char flag_good_position;               // The flag is set when generate the food is at good position
unsigned char index = 0;                        // A global variable


void setup(void)
{
  /*  init the buttons */
    pinMode(BUTTON_MIDDLE, INPUT);
    pinMode(BUTTON_UP, INPUT);
    pinMode(BUTTON_DOWN, INPUT);
    pinMode(BUTTON_LEFT, INPUT);
    pinMode(BUTTON_RIGHT, INPUT);  
  /* init the single direction list */
    for(index = 0; index < SIGLE_DIRECTION_LIST_LENGTH; index++)
    {
        snake[index].x = 0; 
        snake[index].y = 0;   
        snake[index].next = &snake[index+1]; 
    }
    snake[SIGLE_DIRECTION_LIST_LENGTH-1].next = NULL;


  /* init the snake, the original position and length */
    snake[0].x = (AREA_BOUND_START_X + AREA_BOUND_END_X) / 2;
    snake[0].y = (AREA_BOUND_START_Y + AREA_BOUND_END_Y) / 2;
   
    for(index = 1; index < SNAKE_LENGTH_MIN + 1; index++)
    {
        snake[index].x = snake[index-1].x;
        snake[index].y = snake[index].y;
    }
    snake[SNAKE_LENGTH_MIN].next = NULL;

  /* init the pointers */
    p_empty_snake = &snake[SNAKE_LENGTH_MIN + 1];
    p_head_snake = &snake[0];
    p_current_snake = p_head_snake;
    p_tail_snake = &snake[SNAKE_LENGTH_MIN];

    p_food = p_empty_snake;
    p_food->x = snake[0].x + 2;
    p_food->y = snake[0].y + 2;
    p_empty_snake = p_empty_snake->next;

  /* init the snake status */
    snake_status.x_or_y = 'y';
    snake_status.move_step = 1;
    snake_status.life = 1; 

    flag_got_food = 0;  
/*******************************************************
              Start and init the serial port
              Clear the screen
*******************************************************/    

    matrix.start();
    delay(1000);
    
    matrix.change_baudrate(57600);
    delay(200);
    Serial.end();
    Serial.begin(57600);

    matrix.clear_screen(RGB_LED_MATRIX_COLOR_BLACK); /* clear the screen */
    /* Init the bounds, and save to bxGDRAM 0 */
    matrix.line(AREA_BOUND_START_X,  AREA_BOUND_START_Y, AREA_BOUND_END_X, AREA_BOUND_START_Y, RGB_LED_MATRIX_COLOR_RED);
    matrix.line(AREA_BOUND_START_X,  AREA_BOUND_START_Y, AREA_BOUND_START_X, AREA_BOUND_END_Y, RGB_LED_MATRIX_COLOR_RED);
    matrix.line(AREA_BOUND_START_X,  AREA_BOUND_END_Y, AREA_BOUND_END_X, AREA_BOUND_END_Y, RGB_LED_MATRIX_COLOR_RED);
    matrix.line(AREA_BOUND_END_X,  AREA_BOUND_START_Y, AREA_BOUND_END_X, AREA_BOUND_END_Y, RGB_LED_MATRIX_COLOR_RED);
    matrix.copy_GDRAM_to(0); 
    matrix.update_screen();    
}


void loop(void)
{
    //matrix.display_bxGDRAM(0);
    /* Read the buttons */
    if(digitalRead(BUTTON_MIDDLE) == 0)
    {
        snake_status.x_or_y = 0;
    }
    else if(digitalRead(BUTTON_UP) == 0)
    {
        if(snake_status.x_or_y != 'y')
        {
            snake_status.x_or_y = 'y';
            snake_status.move_step = -1;
        }    
    }
    else if(digitalRead(BUTTON_DOWN) == 0)
    {
        if(snake_status.x_or_y != 'y')
        {
            snake_status.x_or_y = 'y';
            snake_status.move_step = 1;
        }    
    }
    else if(digitalRead(BUTTON_LEFT) == 0)
    {
        if(snake_status.x_or_y != 'x')
        {
            snake_status.x_or_y = 'x';
            snake_status.move_step = -1;
        }    
    }
    else if(digitalRead(BUTTON_RIGHT) == 0)
    {
        if(snake_status.x_or_y != 'x')
        {
            snake_status.x_or_y = 'x';
            snake_status.move_step = 1;
        }
    }

    /* put the food on the screen */
    matrix.dot(p_food->x, p_food->y, RGB_LED_MATRIX_COLOR_BLUE);
    /* Got the food? */
    if(snake_status.x_or_y == 'x')
    {   
        if(p_head_snake->y == p_food->y)
        {
            if((p_head_snake->x + snake_status.move_step) == p_food->x)
            {
               flag_got_food = 1;     
            }
        }    
    }
    else if (snake_status.x_or_y == 'y')
    {
        if(p_head_snake->x == p_food->x)
        {
            if((p_head_snake->y + snake_status.move_step) == p_food->y)
            {               
               flag_got_food = 1;                   
            }
        }    
    }

    /* Yes, got the food. */
    if (flag_got_food == 1)
    {
       /* add the food to the snake,and get a new one from empty list */
        p_food->next = p_head_snake;
        p_head_snake = p_food;
        flag_got_food = 0;
        {// generate the food, make sure the food's position is good enough
            p_food = p_empty_snake;
            p_empty_snake = p_empty_snake->next;
                   
            do
            {
                flag_good_position = 1;
                //p_food->x = AREA_BOUND_START_X + 1 + (CNT % (AREA_BOUND_END_X - AREA_BOUND_START_X - 1));
                //p_food->y = AREA_BOUND_START_Y + 1 + (CNT % (AREA_BOUND_END_Y - AREA_BOUND_START_Y - 1));
                p_food->x = random( AREA_BOUND_START_X + 1,  AREA_BOUND_END_X);
                p_food->y = random( AREA_BOUND_START_Y + 1,  AREA_BOUND_END_Y);
                
                p_current_snake = p_head_snake;          

            
                while(p_current_snake->next != NULL)
                {
                    if(p_food->x == p_current_snake->x && p_food->y == p_current_snake->y)
                    {
                        flag_good_position = 0;
                    }
                    p_current_snake = p_current_snake->next;        
                }
            }while(flag_good_position == 0);
        }                       
    }
    else        
    {// move the snake head, only the displaying of the head and tail need to be updated
        matrix.dot(p_tail_snake->x, p_tail_snake->y, RGB_LED_MATRIX_COLOR_BLACK); // clear the tail
        p_current_snake = p_head_snake;            
                 
        temp0_snake.x = p_current_snake->x;                // save the head position;
        temp0_snake.y = p_current_snake->y; 
        {
            if(snake_status.x_or_y == 'x')
            {  
                    p_current_snake->x += snake_status.move_step; 
            }
            else if(snake_status.x_or_y == 'y')
            {
                    p_current_snake->y += snake_status.move_step;
            } 
        }
        /* check if the snake has crash */
        if((p_current_snake->x < (AREA_BOUND_START_X + 1)) || (p_current_snake->x > (AREA_BOUND_END_X - 1)) || (p_current_snake->y < (AREA_BOUND_START_Y + 1)) || (p_current_snake->y > (AREA_BOUND_END_Y - 1)))
        {
                snake_status.life = 0;
                snake_status.x_or_y = 0;
        }
        p_current_snake = p_current_snake->next;            // point to next part[B]
                        
        while(p_current_snake->next != NULL)
        {
            temp1_snake.x = p_current_snake->x;             // save current part value[B]
            temp1_snake.y = p_current_snake->y;
            
            p_current_snake->x = temp0_snake.x;             // restore the previous value[A] to current [B]
            p_current_snake->y = temp0_snake.y;
            p_current_snake = p_current_snake->next;        // point to next part[C]
            
            temp0_snake.x = temp1_snake.x;                  // save current value [C]
            temp0_snake.y = temp1_snake.y;              
        }
        p_current_snake->x = temp0_snake.x;             // deal with the tail
        p_current_snake->y = temp0_snake.y; 
    }     
    /* Update the snake head and tail */     
    p_current_snake = p_head_snake;            
    matrix.dot(p_current_snake->x, p_current_snake->y, RGB_LED_MATRIX_COLOR_GREEN); 
    while(p_current_snake->next != NULL)
    {
        p_current_snake = p_current_snake->next;
    }
    matrix.dot(p_current_snake->x, p_current_snake->y, RGB_LED_MATRIX_COLOR_WHITE);

    if(p_food->next == &snake[SNAKE_LENGTH_MAX])
    {
        //matrix.display_customer_bmp1(AREA_BOUND_START_X, AREA_BOUND_START_Y);
        matrix.display_ascii_0508_nbg('W', AREA_BOUND_START_X, AREA_BOUND_START_Y + 4, RGB_LED_MATRIX_COLOR_WHITE);
        matrix.display_ascii_0508_nbg('i', AREA_BOUND_START_X + 5, AREA_BOUND_START_Y + 4, RGB_LED_MATRIX_COLOR_WHITE);
        matrix.display_ascii_0508_nbg('n', AREA_BOUND_START_X + 10, AREA_BOUND_START_Y + 4, RGB_LED_MATRIX_COLOR_WHITE);                  
        snake_status.life = 0;
        snake_status.x_or_y = 0;
        /* Should jump out and end the game. */
    }

    matrix.screen_overlay(0);
    matrix.update_screen(); 		
    delay(70);

    {// should be dead??  Eat itself??      
        p_current_snake = p_head_snake;
        
        temp0_snake.x = p_current_snake->x;
        temp0_snake.y = p_current_snake->y;
        while(p_current_snake->next != NULL)
        {
            p_current_snake = p_current_snake->next;
            if(p_current_snake->x == temp0_snake.x && p_current_snake->y == temp0_snake.y)
            {
                snake_status.life = 0;
                snake_status.x_or_y = 0;
                break;
            }        
        }
    }                       
}


